@@ -10,7 +10,6 @@ import (
 	"fmt"
 	"io"
 	"io/fs"
-	"io/ioutil"
 	"net/url"
 	"os"
 	"os/exec"
@@ -37,7 +36,7 @@ var (
 const (
 	ErrorSuccessRebootInitiated = 1641
 	ErrorSuccessRebootRequired  = 3010
-	currentMachineVersion       = 2
+	currentMachineVersion       = 3
 )
 
 const containersConf = `[containers]
@@ -46,6 +45,9 @@ const containersConf = `[containers]
 cgroup_manager = "cgroupfs"
 `
 
+const registriesConf = `unqualified-search-registries=["docker.io"]
+`
+
 const appendPort = `grep -q Port\ %d /etc/ssh/sshd_config || echo Port %d >> /etc/ssh/sshd_config`
 
 const configServices = `ln -fs /usr/lib/systemd/system/sshd.service /etc/systemd/system/multi-user.target.wants/sshd.service
@@ -265,11 +267,16 @@ func (p *Provider) NewMachine(opts machine.InitOptions) (machine.VM, error) {
 }
 
 func getConfigPath(name string) (string, error) {
+	return getConfigPathExt(name, "json")
+}
+
+func getConfigPathExt(name string, extension string) (string, error) {
 	vmConfigDir, err := machine.GetConfDir(vmtype)
 	if err != nil {
 		return "", err
 	}
-	return filepath.Join(vmConfigDir, name+".json"), nil
+
+	return filepath.Join(vmConfigDir, fmt.Sprintf("%s.%s", name, extension)), nil
 }
 
 // LoadByName reads a json file that describes a known qemu vm
@@ -315,6 +322,11 @@ func (v *MachineVM) migrateMachine(configPath string) error {
 		return err
 	}
 
+	// Update older machines missing unqualified search config
+	if err := configureRegistries(v, toDist(v.Name)); err != nil {
+		return err
+	}
+
 	v.Version = currentMachineVersion
 	return v.writeConfig()
 }
@@ -417,14 +429,24 @@ func downloadDistro(v *MachineVM, opts machine.InitOptions) error {
 }
 
 func (v *MachineVM) writeConfig() error {
+	const format = "could not write machine json config: %w"
 	jsonFile := v.ConfigPath
-
-	b, err := json.MarshalIndent(v, "", " ")
+	tmpFile, err := getConfigPathExt(v.Name, "tmp")
 	if err != nil {
 		return err
 	}
-	if err := ioutil.WriteFile(jsonFile, b, 0644); err != nil {
-		return fmt.Errorf("could not write machine json config: %w", err)
+
+	b, err := json.MarshalIndent(v, "", " ")
+	if err != nil {
+		return fmt.Errorf(format, err)
+	}
+
+	if err := os.WriteFile(tmpFile, b, 0644); err != nil {
+		return fmt.Errorf(format, err)
+	}
+
+	if err := os.Rename(tmpFile, jsonFile); err != nil {
+		return fmt.Errorf(format, err)
 	}
 
 	return nil
@@ -472,7 +494,7 @@ func provisionWSLDist(v *MachineVM) (string, error) {
 	}
 
 	// Fixes newuidmap
-	if err = wslInvoke(dist, "rpm", "-q", "--restore", "shadow-utils", "2>/dev/null"); err != nil {
+	if err = wslInvoke(dist, "rpm", "--restore", "shadow-utils"); err != nil {
 		return "", fmt.Errorf("package permissions restore of shadow-utils on guest OS failed: %w", err)
 	}
 
@@ -555,6 +577,10 @@ func configureSystem(v *MachineVM, dist string) error {
 		return fmt.Errorf("could not create containers.conf for guest OS: %w", err)
 	}
 
+	if err := configureRegistries(v, dist); err != nil {
+		return err
+	}
+
 	if err := wslInvoke(dist, "sh", "-c", "echo wsl > /etc/containers/podman-machine"); err != nil {
 		return fmt.Errorf("could not create podman-machine file for guest OS: %w", err)
 	}
@@ -566,7 +592,7 @@ func configureSystem(v *MachineVM, dist string) error {
 	return nil
 }
 
-func configureProxy(dist string, useProxy bool) error {
+func configureProxy(dist string, useProxy bool, quiet bool) error {
 	if !useProxy {
 		_ = wslInvoke(dist, "sh", "-c", clearProxySettings)
 		return nil
@@ -587,8 +613,9 @@ func configureProxy(dist string, useProxy bool) error {
 		if exitErr, isExit := err.(*exec.ExitError); isExit && exitErr.ExitCode() != 42 {
 			return fmt.Errorf("%v: %w", failMessage, err)
 		}
-
-		fmt.Println("Installing proxy support")
+		if !quiet {
+			fmt.Println("Installing proxy support")
+		}
 		_ = wslPipe(proxyConfigSetup, dist, "sh", "-c",
 			"cat > /usr/local/bin/proxyinit; chmod 755 /usr/local/bin/proxyinit")
 
@@ -609,6 +636,15 @@ func enableUserLinger(v *MachineVM, dist string) error {
 	return nil
 }
 
+func configureRegistries(v *MachineVM, dist string) error {
+	cmd := "cat > /etc/containers/registries.conf.d/999-podman-machine.conf"
+	if err := wslPipe(registriesConf, dist, "sh", "-c", cmd); err != nil {
+		return fmt.Errorf("could not configure registries on guest OS: %w", err)
+	}
+
+	return nil
+}
+
 func installScripts(dist string) error {
 	if err := wslPipe(enterns, dist, "sh", "-c",
 		"cat > /usr/local/bin/enterns; chmod 755 /usr/local/bin/enterns"); err != nil {
@@ -638,13 +674,13 @@ func installScripts(dist string) error {
 }
 
 func checkAndInstallWSL(opts machine.InitOptions) (bool, error) {
-	if isWSLInstalled() {
+	if IsWSLInstalled() {
 		return true, nil
 	}
 
 	admin := hasAdminRights()
 
-	if !isWSLFeatureEnabled() {
+	if !IsWSLFeatureEnabled() {
 		return false, attemptFeatureInstall(opts, admin)
 	}
 
@@ -939,7 +975,7 @@ func (v *MachineVM) Set(_ string, opts machine.SetOptions) ([]error, error) {
 	if opts.Rootful != nil && v.Rootful != *opts.Rootful {
 		err := v.setRootful(*opts.Rootful)
 		if err != nil {
-			setErrors = append(setErrors, fmt.Errorf("error setting rootful option: %w", err))
+			setErrors = append(setErrors, fmt.Errorf("setting rootful option: %w", err))
 		} else {
 			v.Rootful = *opts.Rootful
 		}
@@ -961,14 +997,14 @@ func (v *MachineVM) Set(_ string, opts machine.SetOptions) ([]error, error) {
 	return setErrors, v.writeConfig()
 }
 
-func (v *MachineVM) Start(name string, _ machine.StartOptions) error {
+func (v *MachineVM) Start(name string, opts machine.StartOptions) error {
 	if v.isRunning() {
 		return fmt.Errorf("%q is already running", name)
 	}
 
 	dist := toDist(name)
 	useProxy := setupWslProxyEnv()
-	if err := configureProxy(dist, useProxy); err != nil {
+	if err := configureProxy(dist, useProxy, opts.Quiet); err != nil {
 		return err
 	}
 
@@ -977,7 +1013,7 @@ func (v *MachineVM) Start(name string, _ machine.StartOptions) error {
 		return fmt.Errorf("the WSL bootstrap script failed: %w", err)
 	}
 
-	if !v.Rootful {
+	if !v.Rootful && !opts.NoInfo {
 		fmt.Printf("\nThis machine is currently configured in rootless mode. If your containers\n")
 		fmt.Printf("require root permissions (e.g. ports < 1024), or if you run into compatibility\n")
 		fmt.Printf("issues with non-podman clients, you can switch using the following command: \n")
@@ -990,22 +1026,24 @@ func (v *MachineVM) Start(name string, _ machine.StartOptions) error {
 	}
 
 	globalName, pipeName, err := launchWinProxy(v)
-	if err != nil {
-		fmt.Fprintln(os.Stderr, "API forwarding for Docker API clients is not available due to the following startup failures.")
-		fmt.Fprintf(os.Stderr, "\t%s\n", err.Error())
-		fmt.Fprintln(os.Stderr, "\nPodman clients are still able to connect.")
-	} else {
-		fmt.Printf("API forwarding listening on: %s\n", pipeName)
-		if globalName {
-			fmt.Printf("\nDocker API clients default to this address. You do not need to set DOCKER_HOST.\n")
+	if !opts.NoInfo {
+		if err != nil {
+			fmt.Fprintln(os.Stderr, "API forwarding for Docker API clients is not available due to the following startup failures.")
+			fmt.Fprintf(os.Stderr, "\t%s\n", err.Error())
+			fmt.Fprintln(os.Stderr, "\nPodman clients are still able to connect.")
 		} else {
-			fmt.Printf("\nAnother process was listening on the default Docker API pipe address.\n")
-			fmt.Printf("You can still connect Docker API clients by setting DOCKER HOST using the\n")
-			fmt.Printf("following powershell command in your terminal session:\n")
-			fmt.Printf("\n\t$Env:DOCKER_HOST = '%s'\n", pipeName)
-			fmt.Printf("\nOr in a classic CMD prompt:\n")
-			fmt.Printf("\n\tset DOCKER_HOST = '%s'\n", pipeName)
-			fmt.Printf("\nAlternatively terminate the other process and restart podman machine.\n")
+			fmt.Printf("API forwarding listening on: %s\n", pipeName)
+			if globalName {
+				fmt.Printf("\nDocker API clients default to this address. You do not need to set DOCKER_HOST.\n")
+			} else {
+				fmt.Printf("\nAnother process was listening on the default Docker API pipe address.\n")
+				fmt.Printf("You can still connect Docker API clients by setting DOCKER HOST using the\n")
+				fmt.Printf("following powershell command in your terminal session:\n")
+				fmt.Printf("\n\t$Env:DOCKER_HOST = '%s'\n", pipeName)
+				fmt.Printf("\nOr in a classic CMD prompt:\n")
+				fmt.Printf("\n\tset DOCKER_HOST = '%s'\n", pipeName)
+				fmt.Printf("\nAlternatively terminate the other process and restart podman machine.\n")
+			}
 		}
 	}
 
@@ -1061,7 +1099,7 @@ func launchWinProxy(v *MachineVM) (bool, string, error) {
 		return globalName, "", err
 	}
 
-	return globalName, pipePrefix + waitPipe, waitPipeExists(waitPipe, 30, func() error {
+	return globalName, pipePrefix + waitPipe, waitPipeExists(waitPipe, 80, func() error {
 		active, exitCode := machine.GetProcessState(cmd.Process.Pid)
 		if !active {
 			return fmt.Errorf("win-sshproxy.exe failed to start, exit code: %d (see windows event logs)", exitCode)
@@ -1099,15 +1137,16 @@ func waitPipeExists(pipeName string, retries int, checkFailure func() error) err
 		if fail := checkFailure(); fail != nil {
 			return fail
 		}
-		time.Sleep(100 * time.Millisecond)
+		time.Sleep(250 * time.Millisecond)
 	}
 
 	return err
 }
 
-func isWSLInstalled() bool {
-	cmd := exec.Command("wsl", "--status")
+func IsWSLInstalled() bool {
+	cmd := SilentExecCmd("wsl", "--status")
 	out, err := cmd.StdoutPipe()
+	cmd.Stderr = nil
 	if err != nil {
 		return false
 	}
@@ -1131,13 +1170,12 @@ func isWSLInstalled() bool {
 	return true
 }
 
-func isWSLFeatureEnabled() bool {
-	cmd := exec.Command("wsl", "--set-default-version", "2")
-	return cmd.Run() == nil
+func IsWSLFeatureEnabled() bool {
+	return SilentExec("wsl", "--set-default-version", "2") == nil
 }
 
 func isWSLRunning(dist string) (bool, error) {
-	cmd := exec.Command("wsl", "-l", "--running")
+	cmd := exec.Command("wsl", "-l", "--running", "--quiet")
 	out, err := cmd.StdoutPipe()
 	if err != nil {
 		return false, err
@@ -1285,7 +1323,7 @@ func readWinProxyTid(v *MachineVM) (uint32, uint32, string, error) {
 	}
 
 	tidFile := filepath.Join(stateDir, winSshProxyTid)
-	contents, err := ioutil.ReadFile(tidFile)
+	contents, err := os.ReadFile(tidFile)
 	if err != nil {
 		return 0, 0, "", err
 	}
@@ -1340,7 +1378,7 @@ func (v *MachineVM) Remove(name string, opts machine.RemoveOptions) (string, fun
 			logrus.Error(err)
 		}
 		for _, f := range files {
-			if err := os.RemoveAll(f); err != nil {
+			if err := machine.GuardedRemoveAll(f); err != nil {
 				logrus.Error(err)
 			}
 		}
@@ -1645,7 +1683,7 @@ func (p *Provider) RemoveAndCleanMachines() error {
 		}
 		prevErr = err
 	} else {
-		err := os.RemoveAll(dataDir)
+		err := machine.GuardedRemoveAll(dataDir)
 		if err != nil {
 			if prevErr != nil {
 				logrus.Error(prevErr)
@@ -1662,7 +1700,7 @@ func (p *Provider) RemoveAndCleanMachines() error {
 		}
 		prevErr = err
 	} else {
-		err := os.RemoveAll(confDir)
+		err := machine.GuardedRemoveAll(confDir)
 		if err != nil {
 			if prevErr != nil {
 				logrus.Error(prevErr)
