@@ -295,11 +295,9 @@ func ConvertContainer(container *parser.UnitFile, isUser bool) (*parser.UnitFile
 		"--log-driver", "passthrough",
 	)
 
-	// We use crun as the runtime and delegated groups to it
+	// We delegate groups to the runtime
 	service.Add(ServiceGroup, "Delegate", "yes")
-	podman.add(
-		"--runtime", "/usr/bin/crun",
-		"--cgroups=split")
+	podman.add("--cgroups=split")
 
 	timezone, ok := container.Lookup(ContainerGroup, KeyTimezone)
 	if ok && len(timezone) > 0 {
@@ -745,6 +743,9 @@ func ConvertKube(kube *parser.UnitFile, isUser bool) (*parser.UnitFile, error) {
 
 		// Use a service container
 		"--service-container=true",
+
+		// We want output to the journal, so use the log driver.
+		"--log-driver", "passthrough",
 	)
 
 	if err := handleUserRemap(kube, KubeGroup, execStart, isUser, false); err != nil {
